use std::fmt;

use jsonrpc_core::{Error as JsonRpcError, ErrorCode};
use peach_network::NetworkError;
use peach_stats::StatsError;
use serde_json::Error as SerdeError;

/// Custom error type encapsulating all possible errors for a JSON-RPC server
/// and associated methods.
#[derive(Debug)]
pub enum ServerError {
    /// An error returned from the `peach-network` library.
    Network(NetworkError),
    /// Failed to serialize a data structure.
    Serialize(SerdeError),
    /// An error returned from the `peach-stats` library.
    Stats(StatsError),
    /// An expected JSON-RPC method parameter was not provided.
    MissingParameter(JsonRpcError),
    /// Failed to parse a provided JSON-RPC method parameter.
    ParseParameter(JsonRpcError),
}

impl fmt::Display for ServerError {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        match *self {
            ServerError::ParseParameter(ref source) => {
                write!(f, "Failed to parse parameter: {}", source)
            }
            ServerError::MissingParameter(ref source) => {
                write!(f, "Missing expected parameter: {}", source)
            }
            ServerError::Network(ref source) => {
                write!(f, "{}", source)
            }
            ServerError::Serialize(ref source) => {
                write!(f, "Serde serialization failure: {}", source)
            }
            ServerError::Stats(ref source) => {
                write!(f, "{}", source)
            }
        }
    }
}

impl From<ServerError> for JsonRpcError {
    fn from(err: ServerError) -> Self {
        match &err {
            ServerError::Network(source) => JsonRpcError {
                code: ErrorCode::ServerError(-32001),
                message: format!("{}", source),
                data: None,
            },
            ServerError::Stats(source) => JsonRpcError {
                code: ErrorCode::ServerError(-32003),
                message: format!("{}", source),
                data: None,
            },
            ServerError::Serialize(source) => JsonRpcError {
                code: ErrorCode::ServerError(-32000),
                message: format!("{}", source),
                data: None,
            },
            ServerError::MissingParameter(source) => source.clone(),
            ServerError::ParseParameter(source) => source.clone(),
        }
    }
}
