//! # peach-jsonrpc-server
//!
//! A JSON-RPC server which exposes an API over HTTP.

use std::env;
use std::result::Result;

use jsonrpc_core::{Error as RpcCoreError, IoHandler, Params, Value};
use jsonrpc_http_server::{AccessControlAllowOrigin, DomainsValidation, ServerBuilder};
use log::info;
use peach_network::network;
use peach_stats::stats;

mod error;
mod params;

use crate::error::ServerError;
use crate::params::{Iface, IfaceId, IfaceIdPass, IfaceSsid, WiFi, WlanAndAp};

/// Create JSON-RPC I/O handler, add RPC methods and launch HTTP server.
pub fn run() -> Result<(), ServerError> {
    info!("Starting up.");

    info!("Creating JSON-RPC I/O handler.");
    let mut io = IoHandler::default();

    io.add_method("ping", |_| async {
        Ok(Value::String("success".to_string()))
    });

    // TODO: add blocks of methods according to provided flags

    /* PEACH-NETWORK RPC METHODS */

    // get - all network rpc methods for querying state

    io.add_method("available_networks", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => {
                match network::available_networks(&i.iface).map_err(ServerError::Network)? {
                    Some(list) => {
                        let json_list =
                            serde_json::to_string(&list).map_err(ServerError::Serialize)?;
                        Ok(Value::String(json_list))
                    }
                    // return `Null` if no networks were found
                    None => Ok(Value::Null),
                }
            }
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("id", |params: Params| async move {
        let parsed: Result<IfaceSsid, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::id(&i.iface, &i.ssid).map_err(ServerError::Network)? {
                Some(id) => Ok(Value::String(id)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("ip", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::ip(&i.iface).map_err(ServerError::Network)? {
                Some(ip) => Ok(Value::String(ip)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("rssi", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::rssi(&i.iface).map_err(ServerError::Network)? {
                Some(rssi) => Ok(Value::String(rssi)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("rssi_percent", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::rssi_percent(&i.iface).map_err(ServerError::Network)? {
                Some(rssi) => Ok(Value::String(rssi)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("saved_networks", |_| async {
        let list = network::saved_networks().map_err(ServerError::Network)?;
        match list {
            Some(list) => {
                let json_list = serde_json::to_string(&list).map_err(ServerError::Serialize)?;
                Ok(Value::String(json_list))
            }
            None => Ok(Value::Null),
        }
    });

    io.add_method("ssid", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::ssid(&i.iface).map_err(ServerError::Network)? {
                Some(ip) => Ok(Value::String(ip)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("state", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::state(&i.iface).map_err(ServerError::Network)? {
                Some(state) => Ok(Value::String(state)),
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("status", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::status(&i.iface).map_err(ServerError::Network)? {
                Some(status) => {
                    let json_status =
                        serde_json::to_string(&status).map_err(ServerError::Serialize)?;
                    Ok(Value::String(json_status))
                }
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("traffic", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::traffic(&i.iface).map_err(ServerError::Network)? {
                Some(traffic) => {
                    let json_traffic =
                        serde_json::to_string(&traffic).map_err(ServerError::Serialize)?;
                    Ok(Value::String(json_traffic))
                }
                None => Ok(Value::Null),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    // set - all network rpc methods for modifying state

    io.add_method("add", |params: Params| async move {
        let parsed: Result<WiFi, RpcCoreError> = params.parse();
        match parsed {
            Ok(w) => match network::add(&w.iface, &w.ssid, &w.pass) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("check_iface", |params: Params| async move {
        let parsed: Result<WlanAndAp, RpcCoreError> = params.parse();
        match parsed {
            Ok(w) => match network::check_iface(&w.wlan_iface, &w.ap_iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("connect", |params: Params| async move {
        let parsed: Result<IfaceId, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::connect(&i.id, &i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("delete", |params: Params| async move {
        let parsed: Result<IfaceId, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::delete(&i.id, &i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("disable", |params: Params| async move {
        let parsed: Result<IfaceId, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::disable(&i.id, &i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("disconnect", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::disconnect(&i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("modify", |params: Params| async move {
        let parsed: Result<IfaceIdPass, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::modify(&i.iface, &i.id, &i.pass) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("reassociate", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::reassociate(&i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("reconfigure", |_| async {
        match network::reconfigure() {
            Ok(_) => Ok(Value::String("success".to_string())),
            Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
        }
    });

    io.add_method("reconnect", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::reconnect(&i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    io.add_method("save", |_| async {
        match network::save() {
            Ok(_) => Ok(Value::String("success".to_string())),
            Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
        }
    });

    io.add_method("start_iface_service", |params: Params| async move {
        let parsed: Result<Iface, RpcCoreError> = params.parse();
        match parsed {
            Ok(i) => match network::start_iface_service(&i.iface) {
                Ok(_) => Ok(Value::String("success".to_string())),
                Err(e) => Err(RpcCoreError::from(ServerError::Network(e))),
            },
            Err(e) => Err(RpcCoreError::from(ServerError::MissingParameter(e))),
        }
    });

    /* PEACH-STATS RPC METHODS */

    io.add_method("cpu_stats", |_| async {
        info!("Fetching CPU statistics.");
        let cpu = stats::cpu_stats().map_err(ServerError::Stats)?;
        let json_cpu = serde_json::to_string(&cpu).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_cpu))
    });

    io.add_method("cpu_stats_percent", |_| async {
        info!("Fetching CPU statistics as percentages.");
        let cpu = stats::cpu_stats_percent().map_err(ServerError::Stats)?;
        let json_cpu = serde_json::to_string(&cpu).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_cpu))
    });

    io.add_method("disk_usage", |_| async {
        info!("Fetching disk usage statistics.");
        let disks = stats::disk_usage().map_err(ServerError::Stats)?;
        let json_disks = serde_json::to_string(&disks).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_disks))
    });

    io.add_method("load_average", |_| async {
        info!("Fetching system load average statistics.");
        let avg = stats::load_average().map_err(ServerError::Stats)?;
        let json_avg = serde_json::to_string(&avg).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_avg))
    });

    io.add_method("mem_stats", |_| async {
        info!("Fetching current memory statistics.");
        let mem = stats::mem_stats().map_err(ServerError::Stats)?;
        let json_mem = serde_json::to_string(&mem).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_mem))
    });

    io.add_method("uptime", |_| async {
        info!("Fetching system uptime.");
        let uptime = stats::uptime().map_err(ServerError::Stats)?;
        let json_uptime = serde_json::to_string(&uptime).map_err(ServerError::Serialize)?;

        Ok(Value::String(json_uptime))
    });

    let http_server =
        env::var("PEACH_JSONRPC_SERVER").unwrap_or_else(|_| "127.0.0.1:5110".to_string());

    info!("Starting JSON-RPC server on {}.", http_server);
    let server = ServerBuilder::new(io)
        .cors(DomainsValidation::AllowOnly(vec![
            AccessControlAllowOrigin::Null,
        ]))
        .start_http(
            &http_server
                .parse()
                .expect("Invalid HTTP address and port combination"),
        )
        .expect("Unable to start RPC server");

    server.wait();

    Ok(())
}

#[cfg(test)]
mod tests {
    use super::*;

    use jsonrpc_core::{Error as JsonRpcError, ErrorCode};
    use jsonrpc_test as test_rpc;

    #[test]
    fn rpc_success() {
        let rpc = {
            let mut io = IoHandler::new();
            io.add_method("rpc_success_response", |_| async {
                Ok(Value::String("success".into()))
            });
            test_rpc::Rpc::from(io)
        };

        assert_eq!(rpc.request("rpc_success_response", &()), r#""success""#);
    }

    #[test]
    fn rpc_parse_error() {
        let rpc = {
            let mut io = IoHandler::new();
            io.add_method("rpc_parse_error", |_| async {
                let e = JsonRpcError {
                    code: ErrorCode::ParseError,
                    message: String::from("Parse error"),
                    data: None,
                };
                Err(JsonRpcError::from(ServerError::MissingParameter(e)))
            });
            test_rpc::Rpc::from(io)
        };

        assert_eq!(
            rpc.request("rpc_parse_error", &()),
            r#"{
  "code": -32700,
  "message": "Parse error"
}"#
        );
    }
}
